"""Take our package compat dictionary and give things a pretty format.

```json
{
	name: str
	version: str
	namever: str
	size: int
	home_page: str
	author: str
	license: str
	license_compat: bool
}
```

Formats

- markdown
- json
- csv
- ansi
"""
from __future__ import annotations

from csv import writer
from io import StringIO
from json import dumps

from licensecheck.types import PackageCompat

logger = None
try:
	from metprint import Logger, LogType

	logger = Logger()
except ModuleNotFoundError:
	pass


def markdown(packages: list[PackageCompat], heading: str | None = None) -> str:
	"""Format to Markdown.

	Args:
		packages (list[PackageCompat]): PackageCompats to format
		heading (str, optional): Optional heading to include. Defaults to None.

	Returns:
		str: String to write to a file of stdout
	"""
	if len(packages) == 0:
		return "No packages"

	heading = heading if heading is not None else "# Packages\nFind a list of packages below"
	strBuf = [heading]
	packages = sorted(packages, key=lambda i: i["name"])

	# Summary Table
	strBuf.append("")
	strBuf.append("|Compatible|Package|\n|:--|:--|")
	for pkg in packages:
		strBuf.append(f"|{pkg['license_compat']}|{pkg['name']}|")
	strBuf.append("")

	# Details
	for pkg in packages:
		strBuf.extend(
			[
				f"## {pkg['namever']}",
				f"\n\n- HomePage: {pkg['home_page']}",
				f"\n- Author: {pkg['author']}",
				f"\n- License: {pkg['license']}",
				f"\n- Compatible: {pkg['license_compat']}",
				f"\n- Size: {pkg['size']}",
			]
		)
	return "\n".join(strBuf) + "\n"


def json(packages: list[PackageCompat], heading: str | None = None) -> str:
	"""Format to Json.

	Args:
		packages (list[PackageCompat]): PackageCompats to format
		heading (str, optional): Optional heading to include. Defaults to None.

	Returns:
		str: String to write to a file of stdout
	"""
	packages = sorted(packages, key=lambda i: i["name"])
	out = {
		"heading": (
			heading if heading is not None else "# Packages - Find a list of packages below"
		),
		"packages": packages,
	}
	return dumps(out, indent="\t")


def csv(packages: list[PackageCompat], heading: str | None = None) -> str:
	"""Format to CSV.

	Args:
		packages (list[PackageCompat]): PackageCompats to format
		heading (str, optional): Optional heading to include. Defaults to None.

	Returns:
		str: String to write to a file of stdout
	"""
	packages = sorted(packages, key=lambda i: i["name"])
	output = StringIO()
	csvString = writer(output)
	csvString.writerow(
		[
			(
				heading
				if heading is not None
				else "# Packages - Find a list of packages below "
				"(you may want to delete this line)"
			)
		]
	)
	csvString.writerow(
		["name", "version", "namever", "home_page", "author", "size", "license", "license_compat"]
	)
	for pkg in packages:
		csvString.writerow(
			[
				pkg["name"],
				pkg["version"],
				pkg["namever"],
				pkg["home_page"],
				pkg["author"],
				pkg["size"],
				pkg["license"],
				pkg["license_compat"],
			]
		)
	return output.getvalue()


def ansi(packages: list[PackageCompat], heading: str | None = None) -> str:
	"""Format to ansi.

	Args:
		packages (list[PackageCompat]): PackageCompats to format
		heading (str, optional): Optional heading to include. Defaults to None.

	Returns:
		str: String to write to a file of stdout
	"""
	# pylint: disable=invalid-name
	BLD = "\033[01m"
	CLS = "\033[00m"
	UL = "\033[04m"
	CB = "\033[36m"
	CG = "\033[32m"

	if len(packages) == 0:
		return f"{BLD}{UL}{CB}No packages{CLS}"

	# pylint: enable=invalid-name
	heading = (
		heading
		if heading is not None
		else f"{BLD}{UL}{CB}PackageCompats{CLS}\n\nFind a list of packages below\n"
	)
	strBuf = [heading]
	packages = sorted(packages, key=lambda i: i["name"])

	# Summary Table
	strBuf.append(f"┌{'─'*10}┬{'─'*30}┐")
	strBuf.append("│Compatible│Package                       │")
	strBuf.append(f"├{'─'*10}┼{'─'*30}┤")
	for pkg in packages:
		strBuf.append(f"│{str(pkg['license_compat']): <10}│{pkg['name'][:30]: <30}│")
	strBuf.append(f"└{'─'*10}┴{'─'*30}┘")
	strBuf.append("")

	# Details
	for pkg in packages:
		strBuf.extend(
			[
				f"{BLD}{UL}{CG}{pkg['namever']}{CLS}",
				f"HomePage: {pkg['home_page']}",
				f"Author: {pkg['author']}",
				f"License: {pkg['license']}",
				f"Compatible: {pkg['license_compat']}",
				f"Size: {pkg['size']}\n",
			]
		)
	return "\n".join(strBuf)


def simple(packages: list[PackageCompat]) -> str:
	"""Format to plain text.

	Args:
		packages (list[PackageCompat]): PackageCompats to format

	Returns:
		str: String to write to a file of stdout
	"""
	packages = sorted(packages, key=lambda i: i["name"])

	# Summary Table
	if logger is None:
		strBuf = [f"┌{'─'*10}┬{'─'*20}┬{'─'*20}┐"]
		strBuf.append("│Compatible│Package             │License             │")
		strBuf.append(f"├{'─'*10}┼{'─'*20}┼{'─'*20}┤")
		for pkg in packages:
			strBuf.append(
				f"│{str(pkg['license_compat']): <10}│{pkg['name'][:20]: <20}│{pkg['license'][:20]: <20}│"
			)
		strBuf.append(f"└{'─'*10}┴{'─'*20}┴{'─'*20}┘")
	else:
		strBuf = [logger.logString(f"┌{'─'*20}┬{'─'*30}┐", LogType.NONE, True)]
		strBuf.append(
			logger.logString(
				"│Package             │License                       │", LogType.NONE, True
			)
		)
		strBuf.append(logger.logString(f"├{'─'*20}┼{'─'*30}┤", LogType.NONE, True))
		for pkg in packages:
			strBuf.append(
				logger.logString(
					f"│{pkg['name'][:20]: <20}│{pkg['license'][:30]: <30}│",
					LogType.SUCCESS if pkg["license_compat"] else LogType.ERROR,
				)
			)
		strBuf.append(logger.logString(f"└{'─'*20}┴{'─'*30}┘", LogType.NONE, True))
	return "\n".join(strBuf)


formatMap = {
	"json": json,
	"markdown": markdown,
	"csv": csv,
	"ansi": ansi,
	"simple": simple,
}



"""Get a list of packages with package compatibility.
"""
from __future__ import annotations

import subprocess
import warnings

import requirements

from licensecheck import license_matrix, packageinfo
from licensecheck.types import PackageCompat

USINGS = ["requirements", "poetry"]


def _doSysExec(command: str) -> tuple[int, str]:
	"""Execute a command and check for errors.

	Args:
		command (str): commands as a string

	Raises:
		RuntimeWarning: throw a warning should there be a non exit code

	Returns:
		tuple[int, str]: exit code and message
	"""
	with subprocess.Popen(
		command,
		shell=True,
		stdout=subprocess.PIPE,
		stderr=subprocess.STDOUT,
		encoding="utf-8",
		errors="ignore",
	) as process:
		out = process.communicate()[0]
		exitCode = process.returncode
	return exitCode, out


def getReqs(using: str) -> list[str]:
	"""Get requirements for the end user project/ lib.

	Args:
		using (str): use requirements or poetry

	Returns:
		list[str]: list of requirement packages
	"""
	_ = using.split(":", 1)
	using, extras = _[0], _[1] if len(_) > 1 else None
	if using not in USINGS:
		using = "poetry"
	reqs = set()

	# Is poetry installed?
	if using == "poetry" and _doSysExec("poetry -h")[0] != 0:
		using = "requirements"  # Poetry not installed - fall back to requirements
		warnings.warn(RuntimeWarning("poetry is not on the system path"))

	# Poetry
	if using == "poetry":  # Use poetry show to get dependents of dependencies
		lines = _doSysExec("poetry show " + ("" if extras else "--no-dev"))[1].splitlines(False)
		for line in lines:
			try:
				parts = line.split()
				reqs.add(parts[0])
			except IndexError:
				print(
					"An error occurred with poetry, try running 'poetry show' to "
					"see what went wrong! - (fall back to requirements)"
				)
				using = "requirements"  # Poetry died - fall back to requirements
				break

	# Requirements
	if using == "requirements":
		for reqTxt in (extras or "requirements.txt").split(";"):
			with open(reqTxt, encoding="utf-8") as requirementsTxt:
				for req in requirements.parse(requirementsTxt):
					reqs.add(req.name)
	return list(reqs)


def getDepsWLicenses(
	using: str,
	ignorePackages: list[str],
	failPackages: list[str],
	ignoreLicenses: list[str],
	failLicenses: list[str],
) -> list[PackageCompat]:
	"""Get a list of dependencies with licenses and determin license compatibility.

	Args:
		using (str): use requirements or poetry
		ignorePackages (list[str]): a list of packages to ignore (compat=True)
		failPackages (list[str]): a list of packages to fail (compat=False)
		ignoreLicenses (list[str]): a list of licenses to ignore (skipped, compat may still be False)
		failLicenses (list[str]): a list of licenses to fail (compat=False)

	Returns:
		list[PackageCompat]: list of packagecompat types: dependency info + licence compat
	"""
	reqs = getReqs(using)

	# Get my license
	myLiceTxt = packageinfo.getMyPackageLicense()
	myLice = license_matrix.licenseType(myLiceTxt)[0]

	# Check it is compatible with packages and add a note
	dependenciesWLicenses = []
	for package in packageinfo.getPackages(reqs):
		# Deal with --ignore-packages and --fail-packages
		licenseCompat = False
		if package["name"].lower() in [x.lower() for x in ignorePackages]:
			licenseCompat = True
		elif package["name"].lower() in [x.lower() for x in failPackages]:
			pass  # licenseCompat=False
		# Old behaviour
		else:
			licenseCompat = license_matrix.depCompatWMyLice(
				myLice,
				license_matrix.licenseType(package["license"]),
				license_matrix.licenseType(", ".join(ignoreLicenses)),
				license_matrix.licenseType(", ".join(failLicenses)),
			)
		# Add to list of dependenciesWLicenses
		dependenciesWLicenses.append(PackageCompat(**package, license_compat=licenseCompat))
	return dependenciesWLicenses



"""Get information for installed and online packages.
"""
from __future__ import annotations

from pathlib import Path
from typing import cast

import requests
import tomlkit
from pip._internal.metadata import get_default_environment as pipenv
from pip._internal.metadata.base import BaseDistribution

from licensecheck.types import PackageInfo

UNKNOWN = "UNKNOWN"


def getPackagesFromLocal(requirements: list[str]) -> list[PackageInfo]:
	"""Get a list of package info from local files including version, author
	and	the license.

	Args:
		requirements (list[str]): [description]

	Returns:
		list[PackageInfo]: [description]
	"""
	# Filter our packages
	pkgs = [pkg for pkg in pipenv().iter_distributions() if pkg.canonical_name in requirements]
	pkgInfo = []
	for pkg in pkgs:
		# Get pkg metadata: license, homepage + author
		pkgMetadata = pkg.metadata
		lice = licenseFromClassifierlist(
			[val for key, val in pkgMetadata.items() if key == "Classifier"]
		)
		if lice == UNKNOWN:
			lice = pkgMetadata.get("License", UNKNOWN)
		homePage = pkgMetadata.get("Home-page", UNKNOWN)
		author = pkgMetadata.get("Author", UNKNOWN)
		# append to pkgInfo
		pkgInfo.append(
			{
				"name": pkg.canonical_name,
				"version": str(pkg.version),
				"namever": str(pkg),
				"home_page": homePage,
				"author": author,
				"size": getModuleSize(pkg),
				"license": lice,
			}
		)
	return pkgInfo


def packageInfoFromPypi(requirements: list[str]) -> list[PackageInfo]:
	"""Get a list of package info from pypi.org including version, author
	and	the license.

	Args:
		requirements (list[str]): [description]

	Returns:
		list[PackageInfo]: [description]
	"""
	pkgInfo = []
	for pkg in requirements:
		request = requests.get(f"https://pypi.org/pypi/{pkg}/json")
		response = request.json()
		info = response["info"]
		licenseClassifier = licenseFromClassifierlist(info["classifiers"])
		pkgInfo.append(
			{
				"name": pkg,
				"version": info["version"],
				"namever": f"{pkg} {info['version']}",
				"home_page": info["home_page"],
				"author": info["author"],
				"size": int(response["urls"][-1]["size"]),
				"license": licenseClassifier if licenseClassifier != UNKNOWN else info["license"],
			}
		)
	return pkgInfo


def licenseFromClassifierlist(classifiers: list[str]) -> str:
	"""Get license string from a list of project classifiers.

	Args:
		classifiers (list[str]): list of classifiers

	Returns:
		str: the license name
	"""
	licenses = []
	for val in classifiers:
		if val.startswith("License"):
			lice = val.split(" :: ")[-1]
			if lice != "OSI Approved":
				licenses.append(lice)
	return ", ".join(licenses) if len(licenses) > 0 else UNKNOWN


def getPackages(reqs: list[str]) -> list[PackageInfo]:
	"""Get dependency info.

	Args:
		reqs (list[str]): list of dependency names to gather info on

	Returns:
		list[PackageInfo]: list of dependencies
	"""
	localReqs = getPackagesFromLocal(reqs)
	for localReq in localReqs:
		reqs.remove(localReq["name"])
	onlineReqs = packageInfoFromPypi(reqs)
	return localReqs + onlineReqs


def getMyPackageLicense() -> str:
	"""Get the pyproject data.

	Returns:
		str: license name

	Raises:
		RuntimeError: Must specify a license using license spdx or classifier (tool.poetry or tool.flit)
	"""
	try:
		pyproject = cast(
			dict, tomlkit.api.loads(Path("pyproject.toml").read_text(encoding="utf-8"))
		)
	except FileNotFoundError:
		return input("Enter the project license")
	tool = pyproject["tool"]
	metaData = {"classifiers": [], "license": ""}
	if "poetry" in tool:
		metaData = tool["poetry"]
	elif "flit" in tool:
		metaData = tool["flit"]["metadata"]
	else:
		return input("Enter the project license")
	licenseClassifier = licenseFromClassifierlist(metaData["classifiers"])
	if licenseClassifier != UNKNOWN:
		return licenseClassifier
	if "license" in metaData:
		return str(metaData["license"])
	raise RuntimeError(
		"Must specify a license using license spdx or classifier (tool.poetry or tool.flit)"
	)


def getModuleSize(pkg: BaseDistribution) -> int:
	"""Get the size of a given module as an int.

	Args:
		pkg (BaseDistribution): package to get the size of

	Returns:
		int: size in bytes
	"""
	size = sum(
		f.stat().st_size
		for f in Path(f"{pkg.location}/{pkg.canonical_name.replace('-', '_')}").glob("**/*")
		if f.is_file() and "__pycache__" not in str(f)
	)
	if size > 0:
		return size
	request = requests.get(f"https://pypi.org/pypi/{pkg.canonical_name}/json")
	response = request.json()
	return int(response["urls"][-1]["size"])
